@@ -26,60 +26,63 @@
 //
 // This should work even with dynamic linking, however, the number of
 // params must be identical, i.e., the "ABI" must match.
+//
 
-#include <wasm.h>
-#include <wasm-builder.h>
 #include <asm_v_wasm.h>
-#include <pass.h>
-#include <wasm-emscripten.h>
 #include <ir/literal-utils.h>
+#include <pass.h>
+#include <wasm-builder.h>
+#include <wasm.h>
 
 namespace wasm {
 
 // This should be enough for everybody. (As described above, we need this
 // to match when dynamically linking, and also dynamic linking is why we
 // can't just detect this automatically in the module we see.)
-static const int NUM_PARAMS = 15;
+static const int NUM_PARAMS = 16;
 
 // Converts a value to the ABI type of i64.
 static Expression* toABI(Expression* value, Module* module) {
   Builder builder(*module);
-  switch (value->type) {
-    case i32: {
+  switch (value->type.getBasic()) {
+    case Type::i32: {
       value = builder.makeUnary(ExtendUInt32, value);
       break;
     }
-    case i64: {
+    case Type::i64: {
       // already good
       break;
     }
-    case f32: {
-      value = builder.makeUnary(
-        ExtendUInt32,
-        builder.makeUnary(ReinterpretFloat32, value)
-      );
+    case Type::f32: {
+      value = builder.makeUnary(ExtendUInt32,
+                                builder.makeUnary(ReinterpretFloat32, value));
       break;
     }
-    case f64: {
+    case Type::f64: {
       value = builder.makeUnary(ReinterpretFloat64, value);
       break;
     }
-    case none: {
+    case Type::v128: {
+      WASM_UNREACHABLE("v128 not implemented yet");
+    }
+    case Type::funcref:
+    case Type::externref:
+    case Type::exnref:
+    case Type::anyref:
+    case Type::eqref:
+    case Type::i31ref: {
+      WASM_UNREACHABLE("reference types cannot be converted to i64");
+    }
+    case Type::none: {
       // the value is none, but we need a value here
-      value = builder.makeSequence(
-        value,
-        LiteralUtils::makeZero(i64, *module)
-      );
+      value =
+        builder.makeSequence(value, LiteralUtils::makeZero(Type::i64, *module));
       break;
     }
-    case unreachable: {
+    case Type::unreachable: {
       // can leave it, the call isn't taken anyhow
       break;
     }
-    default: {
-      // SIMD may be interesting some day
-      WASM_UNREACHABLE();
-    }
   }
   return value;
 }
@@ -87,87 +90,85 @@ static Expression* toABI(Expression* value, Module* module) {
 // Converts a value from the ABI type of i64 to the expected type
 static Expression* fromABI(Expression* value, Type type, Module* module) {
   Builder builder(*module);
-  switch (type) {
-    case i32: {
+  switch (type.getBasic()) {
+    case Type::i32: {
       value = builder.makeUnary(WrapInt64, value);
       break;
     }
-    case i64: {
+    case Type::i64: {
       // already good
       break;
     }
-    case f32: {
-      value = builder.makeUnary(
-        ReinterpretInt32,
-        builder.makeUnary(WrapInt64, value)
-      );
+    case Type::f32: {
+      value = builder.makeUnary(ReinterpretInt32,
+                                builder.makeUnary(WrapInt64, value));
       break;
     }
-    case f64: {
+    case Type::f64: {
       value = builder.makeUnary(ReinterpretInt64, value);
       break;
     }
-    case none: {
+    case Type::v128: {
+      WASM_UNREACHABLE("v128 not implemented yet");
+    }
+    case Type::funcref:
+    case Type::externref:
+    case Type::exnref:
+    case Type::anyref:
+    case Type::eqref:
+    case Type::i31ref: {
+      WASM_UNREACHABLE("reference types cannot be converted from i64");
+    }
+    case Type::none: {
       value = builder.makeDrop(value);
     }
-    case unreachable: {
+    case Type::unreachable: {
       // can leave it, the call isn't taken anyhow
       break;
     }
-    default: {
-      // SIMD may be interesting some day
-      WASM_UNREACHABLE();
-    }
   }
   return value;
 }
 
-struct ParallelFuncCastEmulation : public WalkerPass<PostWalker<ParallelFuncCastEmulation>> {
+struct ParallelFuncCastEmulation
+  : public WalkerPass<PostWalker<ParallelFuncCastEmulation>> {
   bool isFunctionParallel() override { return true; }
 
   Pass* create() override { return new ParallelFuncCastEmulation(ABIType); }
 
-  ParallelFuncCastEmulation(Name ABIType) : ABIType(ABIType) {}
+  ParallelFuncCastEmulation(Signature ABIType) : ABIType(ABIType) {}
 
   void visitCallIndirect(CallIndirect* curr) {
     if (curr->operands.size() > NUM_PARAMS) {
-      Fatal() << "FuncCastEmulation::NUM_PARAMS needs to be at least " <<
-                 curr->operands.size();
+      Fatal() << "FuncCastEmulation::NUM_PARAMS needs to be at least "
+              << curr->operands.size();
     }
     for (Expression*& operand : curr->operands) {
       operand = toABI(operand, getModule());
     }
     // Add extra operands as needed.
     while (curr->operands.size() < NUM_PARAMS) {
-      curr->operands.push_back(LiteralUtils::makeZero(i64, *getModule()));
+      curr->operands.push_back(LiteralUtils::makeZero(Type::i64, *getModule()));
     }
     // Set the new types
-    curr->fullType = ABIType;
+    curr->sig = ABIType;
     auto oldType = curr->type;
-    curr->type = i64;
+    curr->type = Type::i64;
     curr->finalize(); // may be unreachable
     // Fix up return value
     replaceCurrent(fromABI(curr, oldType, getModule()));
   }
 
 private:
-  // the name of a type for a call with the right params and return
-  Name ABIType;
+  // The signature of a call with the right params and return
+  Signature ABIType;
 };
 
 struct FuncCastEmulation : public Pass {
   void run(PassRunner* runner, Module* module) override {
     // we just need the one ABI function type for all indirect calls
-    std::string sig = "j";
-    for (Index i = 0; i < NUM_PARAMS; i++) {
-      sig += 'j';
-    }
-    ABIType = ensureFunctionType(sig, module)->name;
-    // Add a way for JS to call into the table (as our i64 ABI means an i64
-    // is returned when there is a return value, which JS engines will fail on),
-    // using dynCalls
-    EmscriptenGlueGenerator generator(*module);
-    generator.generateDynCallThunks();
+    Signature ABIType(Type(std::vector<Type>(NUM_PARAMS, Type::i64)),
+                      Type::i64);
     // Add a thunk for each function in the table, and do the call through it.
     std::unordered_map<Name, Name> funcThunks;
     for (auto& segment : module->table.segments) {
@@ -183,51 +184,42 @@ struct FuncCastEmulation : public Pass {
       }
     }
     // update call_indirects
-    PassRunner subRunner(module, runner->options);
-    subRunner.setIsNested(true);
-    subRunner.add<ParallelFuncCastEmulation>(ABIType);
-    subRunner.run();
+    ParallelFuncCastEmulation(ABIType).run(runner, module);
   }
 
 private:
-  // the name of a type for a call with the right params and return
-  Name ABIType;
-
   // Creates a thunk for a function, casting args and return value as needed.
   Name makeThunk(Name name, Module* module) {
     Name thunk = std::string("byn$fpcast-emu$") + name.str;
     if (module->getFunctionOrNull(thunk)) {
-      Fatal() << "FuncCastEmulation::makeThunk seems a thunk name already in use. Was the pass already run on this code?";
+      Fatal() << "FuncCastEmulation::makeThunk seems a thunk name already in "
+                 "use. Was the pass already run on this code?";
     }
     // The item in the table may be a function or a function import.
     auto* func = module->getFunction(name);
-    std::vector<Type>& params = func->params;
-    Type type = func->result;
+    Type type = func->sig.results;
     Builder builder(*module);
     std::vector<Expression*> callOperands;
-    for (Index i = 0; i < params.size(); i++) {
-      callOperands.push_back(fromABI(builder.makeGetLocal(i, i64), params[i], module));
+    Index i = 0;
+    for (const auto& param : func->sig.params) {
+      callOperands.push_back(
+        fromABI(builder.makeLocalGet(i++, Type::i64), param, module));
     }
     auto* call = builder.makeCall(name, callOperands, type);
     std::vector<Type> thunkParams;
     for (Index i = 0; i < NUM_PARAMS; i++) {
-      thunkParams.push_back(i64);
+      thunkParams.push_back(Type::i64);
     }
-    auto* thunkFunc = builder.makeFunction(
-      thunk,
-      std::move(thunkParams),
-      i64,
-      {}, // no vars
-      toABI(call, module)
-    );
-    thunkFunc->type = ABIType;
+    auto* thunkFunc =
+      builder.makeFunction(thunk,
+                           Signature(Type(thunkParams), Type::i64),
+                           {}, // no vars
+                           toABI(call, module));
     module->addFunction(thunkFunc);
     return thunk;
   }
 };
 
-Pass* createFuncCastEmulationPass() {
-  return new FuncCastEmulation();
-}
+Pass* createFuncCastEmulationPass() { return new FuncCastEmulation(); }
 
 } // namespace wasm
