@@ -39,106 +39,54 @@
 
 namespace wasm {
 
-static const Name NOTPRECOMPUTABLE_FLOW("Binaryen|notprecomputable");
-
-// Limit evaluation depth for 2 reasons: first, it is highly unlikely
-// that we can do anything useful to precompute a hugely nested expression
-// (we should succed at smaller parts of it first). Second, a low limit is
-// helpful to avoid platform differences in native stack sizes.
-static const Index MAX_DEPTH = 50;
-
-typedef std::unordered_map<LocalGet*, Literal> GetValues;
+typedef std::unordered_map<LocalGet*, Literals> GetValues;
 
 // Precomputes an expression. Errors if we hit anything that can't be
-// precomputed.
+// precomputed. Inherits most of its functionality from
+// ConstantExpressionRunner, which it shares with the C-API, but adds handling
+// of GetValues computed during the precompute pass.
 class PrecomputingExpressionRunner
-  : public ExpressionRunner<PrecomputingExpressionRunner> {
-  Module* module;
+  : public ConstantExpressionRunner<PrecomputingExpressionRunner> {
 
-  // map gets to constant values, if they are known to be constant
+  // Concrete values of gets computed during the pass, which the runner does not
+  // know about since it only records values of sets it visits.
   GetValues& getValues;
 
-  // Whether we are trying to precompute down to an expression (which we can do
-  // on say 5 + 6) or to a value (which we can't do on a local.tee that flows a
-  // 7 through it). When we want to replace the expression, we can only do so
-  // when it has no side effects. When we don't care about replacing the
-  // expression, we just want to know if it will contain a known constant.
-  bool replaceExpression;
+  // Limit evaluation depth for 2 reasons: first, it is highly unlikely
+  // that we can do anything useful to precompute a hugely nested expression
+  // (we should succed at smaller parts of it first). Second, a low limit is
+  // helpful to avoid platform differences in native stack sizes.
+  static const Index MAX_DEPTH = 50;
+
+  // Limit loop iterations since loops might be infinite. Since we are going to
+  // replace the expression and must preserve side effects, we limit this to the
+  // very first iteration because a side effect would be necessary to achieve
+  // more than one iteration before becoming concrete.
+  static const Index MAX_LOOP_ITERATIONS = 1;
 
 public:
   PrecomputingExpressionRunner(Module* module,
                                GetValues& getValues,
                                bool replaceExpression)
-    : ExpressionRunner<PrecomputingExpressionRunner>(MAX_DEPTH), module(module),
-      getValues(getValues), replaceExpression(replaceExpression) {}
+    : ConstantExpressionRunner<PrecomputingExpressionRunner>(
+        module,
+        replaceExpression ? FlagValues::PRESERVE_SIDEEFFECTS
+                          : FlagValues::DEFAULT,
+        MAX_DEPTH,
+        MAX_LOOP_ITERATIONS),
+      getValues(getValues) {}
 
-  struct NonstandaloneException {
-  }; // TODO: use a flow with a special name, as this is likely very slow
-
-  Flow visitLoop(Loop* curr) {
-    // loops might be infinite, so must be careful
-    // but we can't tell if non-infinite, since we don't have state, so loops
-    // are just impossible to optimize for now
-    return Flow(NOTPRECOMPUTABLE_FLOW);
-  }
-
-  Flow visitCall(Call* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitCallIndirect(CallIndirect* curr) {
-    return Flow(NOTPRECOMPUTABLE_FLOW);
-  }
   Flow visitLocalGet(LocalGet* curr) {
     auto iter = getValues.find(curr);
     if (iter != getValues.end()) {
-      auto value = iter->second;
-      if (value.isConcrete()) {
-        return Flow(value);
+      auto values = iter->second;
+      if (values.isConcrete()) {
+        return Flow(values);
       }
     }
-    return Flow(NOTPRECOMPUTABLE_FLOW);
+    return ConstantExpressionRunner<
+      PrecomputingExpressionRunner>::visitLocalGet(curr);
   }
-  Flow visitLocalSet(LocalSet* curr) {
-    // If we don't need to replace the whole expression, see if there
-    // is a value flowing through a tee.
-    if (!replaceExpression) {
-      if (curr->type.isConcrete()) {
-        assert(curr->isTee());
-        return visit(curr->value);
-      }
-    }
-    return Flow(NOTPRECOMPUTABLE_FLOW);
-  }
-  Flow visitGlobalGet(GlobalGet* curr) {
-    auto* global = module->getGlobal(curr->name);
-    if (!global->imported() && !global->mutable_) {
-      return visit(global->init);
-    }
-    return Flow(NOTPRECOMPUTABLE_FLOW);
-  }
-  Flow visitGlobalSet(GlobalSet* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitLoad(Load* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitStore(Store* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitAtomicRMW(AtomicRMW* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitAtomicCmpxchg(AtomicCmpxchg* curr) {
-    return Flow(NOTPRECOMPUTABLE_FLOW);
-  }
-  Flow visitAtomicWait(AtomicWait* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitAtomicNotify(AtomicNotify* curr) {
-    return Flow(NOTPRECOMPUTABLE_FLOW);
-  }
-  Flow visitSIMDLoad(SIMDLoad* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitMemoryInit(MemoryInit* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitDataDrop(DataDrop* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitMemoryCopy(MemoryCopy* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitMemoryFill(MemoryFill* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitHost(Host* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitTry(Try* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitThrow(Throw* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitRethrow(Rethrow* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitBrOnExn(BrOnExn* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitPush(Push* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-  Flow visitPop(Pop* curr) { return Flow(NOTPRECOMPUTABLE_FLOW); }
-
-  void trap(const char* why) override { throw NonstandaloneException(); }
 };
 
 struct Precompute
@@ -174,6 +122,40 @@ struct Precompute
     } while (propagate && worked);
   }
 
+  template<typename T> void reuseConstantNode(T* curr, Flow flow) {
+    if (flow.values.isConcrete()) {
+      // reuse a const / ref.null / ref.func node if there is one
+      if (curr->value && flow.values.size() == 1) {
+        Literal singleValue = flow.getSingleValue();
+        if (singleValue.type.isNumber()) {
+          if (auto* c = curr->value->template dynCast<Const>()) {
+            c->value = singleValue;
+            c->finalize();
+            curr->finalize();
+            return;
+          }
+        } else if (singleValue.isNull()) {
+          if (auto* n = curr->value->template dynCast<RefNull>()) {
+            n->finalize(singleValue.type);
+            curr->finalize();
+            return;
+          }
+        } else if (singleValue.type == Type::funcref) {
+          if (auto* r = curr->value->template dynCast<RefFunc>()) {
+            r->func = singleValue.getFunc();
+            r->finalize();
+            curr->finalize();
+            return;
+          }
+        }
+      }
+      curr->value = flow.getConstExpression(*getModule());
+    } else {
+      curr->value = nullptr;
+    }
+    curr->finalize();
+  }
+
   void visitExpression(Expression* curr) {
     // TODO: if local.get, only replace with a constant if we don't care about
     // size...?
@@ -188,36 +170,23 @@ struct Precompute
     }
     // try to evaluate this into a const
     Flow flow = precomputeExpression(curr);
-    if (flow.value.type.isVector()) {
+    if (flow.getType().hasVector()) {
       return;
     }
     if (flow.breaking()) {
-      if (flow.breakTo == NOTPRECOMPUTABLE_FLOW) {
+      if (flow.breakTo == NONCONSTANT_FLOW) {
         return;
       }
       if (flow.breakTo == RETURN_FLOW) {
         // this expression causes a return. if it's already a return, reuse the
         // node
         if (auto* ret = curr->dynCast<Return>()) {
-          if (flow.value.type != Type::none) {
-            // reuse a const value if there is one
-            if (ret->value) {
-              if (auto* value = ret->value->dynCast<Const>()) {
-                value->value = flow.value;
-                value->finalize();
-                return;
-              }
-            }
-            ret->value = Builder(*getModule()).makeConstExpression(flow.value);
-          } else {
-            ret->value = nullptr;
-          }
+          reuseConstantNode(ret, flow);
         } else {
           Builder builder(*getModule());
-          replaceCurrent(
-            builder.makeReturn(flow.value.type != Type::none
-                                 ? builder.makeConstExpression(flow.value)
-                                 : nullptr));
+          replaceCurrent(builder.makeReturn(
+            flow.values.isConcrete() ? flow.getConstExpression(*getModule())
+                                     : nullptr));
         }
         return;
       }
@@ -226,34 +195,19 @@ struct Precompute
       if (auto* br = curr->dynCast<Break>()) {
         br->name = flow.breakTo;
         br->condition = nullptr;
-        if (flow.value.type != Type::none) {
-          // reuse a const value if there is one
-          if (br->value) {
-            if (auto* value = br->value->dynCast<Const>()) {
-              value->value = flow.value;
-              value->finalize();
-              br->finalize();
-              return;
-            }
-          }
-          br->value = Builder(*getModule()).makeConstExpression(flow.value);
-        } else {
-          br->value = nullptr;
-        }
-        br->finalize();
+        reuseConstantNode(br, flow);
       } else {
         Builder builder(*getModule());
-        replaceCurrent(
-          builder.makeBreak(flow.breakTo,
-                            flow.value.type != Type::none
-                              ? builder.makeConstExpression(flow.value)
-                              : nullptr));
+        replaceCurrent(builder.makeBreak(
+          flow.breakTo,
+          flow.values.isConcrete() ? flow.getConstExpression(*getModule())
+                                   : nullptr));
       }
       return;
     }
     // this was precomputed
-    if (flow.value.type.isConcrete()) {
-      replaceCurrent(Builder(*getModule()).makeConstExpression(flow.value));
+    if (flow.values.isConcrete()) {
+      replaceCurrent(flow.getConstExpression(*getModule()));
       worked = true;
     } else {
       ExpressionManipulator::nop(curr);
@@ -273,8 +227,8 @@ private:
       return PrecomputingExpressionRunner(
                getModule(), getValues, replaceExpression)
         .visit(curr);
-    } catch (PrecomputingExpressionRunner::NonstandaloneException&) {
-      return Flow(NOTPRECOMPUTABLE_FLOW);
+    } catch (PrecomputingExpressionRunner::NonconstantException&) {
+      return Flow(NONCONSTANT_FLOW);
     }
   }
 
@@ -285,14 +239,14 @@ private:
   //  (local.tee (i32.const 1))
   // will have value 1 which we can optimize here, but in precomputeExpression
   // we could not do anything.
-  Literal precomputeValue(Expression* curr) {
+  Literals precomputeValue(Expression* curr) {
     // Note that we set replaceExpression to false, as we just care about
     // the value here.
     Flow flow = precomputeExpression(curr, false /* replaceExpression */);
     if (flow.breaking()) {
-      return Literal();
+      return {};
     }
-    return flow.value;
+    return flow.values;
   }
 
   // Propagates values around. Returns whether we propagated.
@@ -313,7 +267,7 @@ private:
       work.insert(curr);
     }
     // the constant value, or none if not a constant
-    std::unordered_map<LocalSet*, Literal> setValues;
+    std::unordered_map<LocalSet*, Literals> setValues;
     // propagate constant values
     while (!work.empty()) {
       auto iter = work.begin();
@@ -326,29 +280,31 @@ private:
         if (setValues[set].isConcrete()) {
           continue; // already known constant
         }
-        auto value = setValues[set] =
-          precomputeValue(Properties::getFallthrough(set->value));
-        if (value.isConcrete()) {
+        auto values = setValues[set] =
+          precomputeValue(Properties::getFallthrough(
+            set->value, getPassOptions(), getModule()->features));
+        if (values.isConcrete()) {
           for (auto* get : localGraph.setInfluences[set]) {
             work.insert(get);
           }
         }
       } else {
         auto* get = curr->cast<LocalGet>();
-        if (getValues[get].isConcrete()) {
+        if (getValues[get].size() >= 1) {
           continue; // already known constant
         }
         // for this get to have constant value, all sets must agree
-        Literal value;
+        Literals values;
         bool first = true;
         for (auto* set : localGraph.getSetses[get]) {
-          Literal curr;
+          Literals curr;
           if (set == nullptr) {
             if (getFunction()->isVar(get->index)) {
-              curr = Literal::makeZero(getFunction()->getLocalType(get->index));
+              curr =
+                Literal::makeZeros(getFunction()->getLocalType(get->index));
             } else {
               // it's a param, so it's hopeless
-              value = Literal();
+              values = {};
               break;
             }
           } else {
@@ -356,25 +312,25 @@ private:
           }
           if (curr.isNone()) {
             // not a constant, give up
-            value = Literal();
+            values = {};
             break;
           }
           // we found a concrete value. compare with the current one
           if (first) {
-            value = curr; // this is the first
+            values = curr; // this is the first
             first = false;
           } else {
-            if (value != curr) {
+            if (values != curr) {
               // not the same, give up
-              value = Literal();
+              values = {};
               break;
             }
           }
         }
         // we may have found a value
-        if (value.isConcrete()) {
+        if (values.isConcrete()) {
           // we did!
-          getValues[get] = value;
+          getValues[get] = values;
           for (auto* set : localGraph.getInfluences[get]) {
             work.insert(set);
           }
