@@ -21,12 +21,12 @@
 #ifndef wasm_shell_interface_h
 #define wasm_shell_interface_h
 
-#include "shared-constants.h"
 #include "asmjs/shared-constants.h"
-#include "support/name.h"
-#include "wasm.h"
-#include "wasm-interpreter.h"
 #include "ir/module-utils.h"
+#include "shared-constants.h"
+#include "support/name.h"
+#include "wasm-interpreter.h"
+#include "wasm.h"
 
 namespace wasm {
 
@@ -44,15 +44,14 @@ struct ShellExternalInterface : ModuleInstance::ExternalInterface {
   class Memory {
     // Use char because it doesn't run afoul of aliasing rules.
     std::vector<char> memory;
-    template<typename T>
-    static bool aligned(const char* address) {
+    template<typename T> static bool aligned(const char* address) {
       static_assert(!(sizeof(T) & (sizeof(T) - 1)), "must be a power of 2");
       return 0 == (reinterpret_cast<uintptr_t>(address) & (sizeof(T) - 1));
     }
     Memory(Memory&) = delete;
     Memory& operator=(const Memory&) = delete;
 
-   public:
+  public:
     Memory() = default;
     void resize(size_t newSize) {
       // Ensure the smallest allocation is large enough that most allocators
@@ -68,16 +67,14 @@ struct ShellExternalInterface : ModuleInstance::ExternalInterface {
         std::memset(&memory[newSize], 0, minSize - newSize);
       }
     }
-    template<typename T>
-    void set(size_t address, T value) {
+    template<typename T> void set(size_t address, T value) {
       if (aligned<T>(&memory[address])) {
         *reinterpret_cast<T*>(&memory[address]) = value;
       } else {
         std::memcpy(&memory[address], &value, sizeof(T));
       }
     }
-    template<typename T>
-    T get(size_t address) {
+    template<typename T> T get(size_t address) {
       if (aligned<T>(&memory[address])) {
         return *reinterpret_cast<T*>(&memory[address]);
       } else {
@@ -98,34 +95,55 @@ struct ShellExternalInterface : ModuleInstance::ExternalInterface {
     table.resize(wasm.table.initial);
   }
 
-  void importGlobals(std::map<Name, Literal>& globals, Module& wasm) override {
+  void importGlobals(std::map<Name, Literals>& globals, Module& wasm) override {
     // add spectest globals
     ModuleUtils::iterImportedGlobals(wasm, [&](Global* import) {
-      if (import->module == SPECTEST && import->base == GLOBAL) {
-        switch (import->type) {
-          case i32: globals[import->name] = Literal(int32_t(666)); break;
-          case i64: globals[import->name] = Literal(int64_t(666)); break;
-          case f32: globals[import->name] = Literal(float(666.6)); break;
-          case f64: globals[import->name] = Literal(double(666.6)); break;
-          case v128: assert(false && "v128 not implemented yet");
-          case none:
-          case unreachable: WASM_UNREACHABLE();
+      if (import->module == SPECTEST && import->base.startsWith(GLOBAL)) {
+        TODO_SINGLE_COMPOUND(import->type);
+        switch (import->type.getBasic()) {
+          case Type::i32:
+            globals[import->name] = {Literal(int32_t(666))};
+            break;
+          case Type::i64:
+            globals[import->name] = {Literal(int64_t(666))};
+            break;
+          case Type::f32:
+            globals[import->name] = {Literal(float(666.6))};
+            break;
+          case Type::f64:
+            globals[import->name] = {Literal(double(666.6))};
+            break;
+          case Type::v128:
+            assert(false && "v128 not implemented yet");
+          case Type::funcref:
+          case Type::externref:
+          case Type::exnref:
+          case Type::anyref:
+          case Type::eqref:
+            globals[import->name] = {Literal::makeNull(import->type)};
+            break;
+          case Type::i31ref:
+            WASM_UNREACHABLE("TODO: i31ref");
+          case Type::none:
+          case Type::unreachable:
+            WASM_UNREACHABLE("unexpected type");
         }
       }
     });
-    if (wasm.memory.imported() && wasm.memory.module == SPECTEST && wasm.memory.base == MEMORY) {
+    if (wasm.memory.imported() && wasm.memory.module == SPECTEST &&
+        wasm.memory.base == MEMORY) {
       // imported memory has initial 1 and max 2
       wasm.memory.initial = 1;
       wasm.memory.max = 2;
     }
   }
 
-  Literal callImport(Function* import, LiteralList& arguments) override {
-    if (import->module == SPECTEST && import->base == PRINT) {
+  Literals callImport(Function* import, LiteralList& arguments) override {
+    if (import->module == SPECTEST && import->base.startsWith(PRINT)) {
       for (auto argument : arguments) {
-        std::cout << '(' << argument << ')' << '\n';
+        std::cout << argument << " : " << argument.type << '\n';
       }
-      return Literal();
+      return {};
     } else if (import->module == ENV && import->base == EXIT) {
       // XXX hack for torture tests
       std::cout << "exit()\n";
@@ -135,17 +153,31 @@ struct ShellExternalInterface : ModuleInstance::ExternalInterface {
             << import->name.str;
   }
 
-  Literal callTable(Index index, LiteralList& arguments, Type result, ModuleInstance& instance) override {
-    if (index >= table.size()) trap("callTable overflow");
+  Literals callTable(Index index,
+                     Signature sig,
+                     LiteralList& arguments,
+                     Type results,
+                     ModuleInstance& instance) override {
+    if (index >= table.size()) {
+      trap("callTable overflow");
+    }
     auto* func = instance.wasm.getFunctionOrNull(table[index]);
-    if (!func) trap("uninitialized table element");
-    if (func->params.size() != arguments.size()) trap("callIndirect: bad # of arguments");
-    for (size_t i = 0; i < func->params.size(); i++) {
-      if (func->params[i] != arguments[i].type) {
+    if (!func) {
+      trap("uninitialized table element");
+    }
+    if (sig != func->sig) {
+      trap("callIndirect: function signatures don't match");
+    }
+    if (func->sig.params.size() != arguments.size()) {
+      trap("callIndirect: bad # of arguments");
+    }
+    size_t i = 0;
+    for (const auto& param : func->sig.params) {
+      if (!Type::isSubType(arguments[i++].type, param)) {
         trap("callIndirect: bad argument type");
       }
     }
-    if (func->result != result) {
+    if (func->sig.results != results) {
       trap("callIndirect: bad result type");
     }
     if (func->imported()) {
@@ -167,28 +199,42 @@ struct ShellExternalInterface : ModuleInstance::ExternalInterface {
     return memory.get<std::array<uint8_t, 16>>(addr);
   }
 
-  void store8(Address addr, int8_t value) override { memory.set<int8_t>(addr, value); }
-  void store16(Address addr, int16_t value) override { memory.set<int16_t>(addr, value); }
-  void store32(Address addr, int32_t value) override { memory.set<int32_t>(addr, value); }
-  void store64(Address addr, int64_t value) override { memory.set<int64_t>(addr, value); }
+  void store8(Address addr, int8_t value) override {
+    memory.set<int8_t>(addr, value);
+  }
+  void store16(Address addr, int16_t value) override {
+    memory.set<int16_t>(addr, value);
+  }
+  void store32(Address addr, int32_t value) override {
+    memory.set<int32_t>(addr, value);
+  }
+  void store64(Address addr, int64_t value) override {
+    memory.set<int64_t>(addr, value);
+  }
   void store128(Address addr, const std::array<uint8_t, 16>& value) override {
     memory.set<std::array<uint8_t, 16>>(addr, value);
   }
 
-  void tableStore(Address addr, Name entry) override {
-    table[addr] = entry;
-  }
+  void tableStore(Address addr, Name entry) override { table[addr] = entry; }
 
-  void growMemory(Address /*oldSize*/, Address newSize) override {
+  bool growMemory(Address /*oldSize*/, Address newSize) override {
+    // Apply a reasonable limit on memory size, 1GB, to avoid DOS on the
+    // interpreter.
+    if (newSize > 1024 * 1024 * 1024) {
+      return false;
+    }
     memory.resize(newSize);
+    return true;
   }
 
   void trap(const char* why) override {
     std::cout << "[trap " << why << "]\n";
     throw TrapException();
   }
+
+  void throwException(Literal exn) override { throw WasmException(exn); }
 };
 
-}
+} // namespace wasm
 
 #endif // wasm_shell_interface_h
