@@ -17,60 +17,58 @@
 #ifndef wasm_wasm_emscripten_h
 #define wasm_wasm_emscripten_h
 
-#include "wasm.h"
-#include "wasm-builder.h"
 #include "support/file.h"
-
+#include "wasm-builder.h"
+#include "wasm.h"
 
 namespace wasm {
 
+Global* getStackPointerGlobal(Module& wasm);
+
 // Class which modifies a wasm module for use with emscripten. Generates
 // runtime functions and emits metadata.
 class EmscriptenGlueGenerator {
 public:
   EmscriptenGlueGenerator(Module& wasm, Address stackPointerOffset = Address(0))
-    : wasm(wasm),
-      builder(wasm),
-      stackPointerOffset(stackPointerOffset),
-      useStackPointerGlobal(stackPointerOffset == 0) { }
+    : wasm(wasm), builder(wasm), stackPointerOffset(stackPointerOffset),
+      useStackPointerGlobal(stackPointerOffset == 0) {}
 
-  void generateRuntimeFunctions();
-  Function* generateMemoryGrowthFunction();
   Function* generateAssignGOTEntriesFunction();
-  void generateStackInitialization(Address addr);
   void generatePostInstantiateFunction();
 
-  // Create thunks for use with emscripten Runtime.dynCall. Creates one for each
-  // signature in the indirect function table.
-  void generateDynCallThunks();
-
-  // Convert stack pointer access from global.get/global.set to calling save
-  // and restore functions.
-  void replaceStackPointerGlobal();
-
-  std::string generateEmscriptenMetadata(
-    Address staticBump, std::vector<Name> const& initializerFunctions);
+  // Remove the import of a mutable __stack_pointer and instead initialize the
+  // stack pointer from an immutable import.
+  void internalizeStackPointerGlobal();
 
+  std::string
+  generateEmscriptenMetadata(Address staticBump,
+                             std::vector<Name> const& initializerFunctions);
 
   void fixInvokeFunctionNames();
 
+  // clang uses name mangling to rename the argc/argv form of main to
+  // __main_argc_argv.  Emscripten in non-standalone mode expects that function
+  // to be exported as main.  This function renames __main_argc_argv to main
+  // as expected by emscripten.
+  void renameMainArgcArgv();
+
   // Emits the data segments to a file. The file contains data from address base
-  // onwards (we must pass in base, as we can't tell it from the wasm - the first
-  // segment may start after a run of zeros, but we need those zeros in the file).
+  // onwards (we must pass in base, as we can't tell it from the wasm - the
+  // first segment may start after a run of zeros, but we need those zeros in
+  // the file).
   void separateDataSegments(Output* outfile, Address base);
 
+  bool standalone = false;
+  bool sideModule = false;
+  bool minimizeWasmChanges = false;
+  bool noDynCalls = false;
+  bool onlyI64DynCalls = false;
+
 private:
   Module& wasm;
   Builder builder;
   Address stackPointerOffset;
   bool useStackPointerGlobal;
-
-  Global* getStackPointerGlobal();
-  Expression* generateLoadStackPointer();
-  Expression* generateStoreStackPointer(Expression* value);
-  void generateStackSaveFunction();
-  void generateStackAllocFunction();
-  void generateStackRestoreFunction();
 };
 
 } // namespace wasm
