@@ -22,20 +22,22 @@
 #ifndef wasm_wasm_s_parser_h
 #define wasm_wasm_s_parser_h
 
-#include "wasm.h"
 #include "mixed_arena.h"
 #include "parsing.h" // for UniqueNameMapper. TODO: move dependency to cpp file?
+#include "wasm-builder.h"
+#include "wasm.h"
 
 namespace wasm {
 
-class SourceLocation
-{
+class SourceLocation {
 public:
   cashew::IString filename;
   uint32_t line;
   uint32_t column;
-  SourceLocation(cashew::IString filename_, uint32_t line_, uint32_t column_ = 0)
-   : filename(filename_), line(line_), column(column_) {}
+  SourceLocation(cashew::IString filename_,
+                 uint32_t line_,
+                 uint32_t column_ = 0)
+    : filename(filename_), line(line_), column(column_) {}
 };
 
 //
@@ -59,7 +61,7 @@ public:
   bool quoted() const { return isStr() && quoted_; }
 
   size_t line = -1;
-  size_t col  = -1;
+  size_t col = -1;
   // original locations at the start/end of the S-Expression list
   SourceLocation* startLoc = nullptr;
   SourceLocation* endLoc = nullptr;
@@ -67,9 +69,7 @@ public:
   // list methods
   List& list();
   Element* operator[](unsigned i);
-  size_t size() {
-    return list().size();
-  }
+  size_t size() { return list().size(); }
 
   // string methods
   cashew::IString str() const;
@@ -111,20 +111,26 @@ private:
 class SExpressionWasmBuilder {
   Module& wasm;
   MixedArena& allocator;
+  IRProfile profile;
+  std::vector<Signature> signatures;
+  std::unordered_map<std::string, size_t> signatureIndices;
   std::vector<Name> functionNames;
-  std::vector<Name> functionTypeNames;
   std::vector<Name> globalNames;
-  int functionCounter;
+  std::vector<Name> eventNames;
+  int functionCounter = 0;
   int globalCounter = 0;
-  std::map<Name, Type> functionTypes; // we need to know function return types before we parse their contents
+  int eventCounter = 0;
+  // we need to know function return types before we parse their contents
+  std::map<Name, Type> functionTypes;
   std::unordered_map<cashew::IString, Index> debugInfoFileIndices;
 
 public:
   // Assumes control of and modifies the input.
-  SExpressionWasmBuilder(Module& wasm, Element& module, Name* moduleName = nullptr);
+  SExpressionWasmBuilder(Module& wasm, Element& module, IRProfile profile);
 
 private:
-  // pre-parse types and function definitions, so we know function return types before parsing their contents
+  // pre-parse types and function definitions, so we know function return types
+  // before parsing their contents
   void preParseFunctionType(Element& s);
   bool isImport(Element& curr);
   void preParseImports(Element& curr);
@@ -132,40 +138,42 @@ private:
 
   // function parsing state
   std::unique_ptr<Function> currFunction;
-  std::map<Name, Type> currLocalTypes;
-  size_t localIndex; // params and vars
-  size_t otherIndex;
   bool brokeToAutoBlock;
 
   UniqueNameMapper nameMapper;
 
+  Signature getFunctionSignature(Element& s);
   Name getFunctionName(Element& s);
-  Name getFunctionTypeName(Element& s);
   Name getGlobalName(Element& s);
-  void parseStart(Element& s) { wasm.addStart(getFunctionName(*s[1]));}
+  Name getEventName(Element& s);
+  void parseStart(Element& s) { wasm.addStart(getFunctionName(*s[1])); }
 
   // returns the next index in s
   size_t parseFunctionNames(Element& s, Name& name, Name& exportName);
   void parseFunction(Element& s, bool preParseImport = false);
 
-  Type stringToType(cashew::IString str, bool allowError=false, bool prefix=false) {
+  Type stringToType(cashew::IString str,
+                    bool allowError = false,
+                    bool prefix = false) {
     return stringToType(str.str, allowError, prefix);
   }
-  Type stringToType(const char* str, bool allowError=false, bool prefix=false);
+  Type
+  stringToType(const char* str, bool allowError = false, bool prefix = false);
+  HeapType stringToHeapType(cashew::IString str, bool prefix = false) {
+    return stringToHeapType(str.str, prefix);
+  }
+  HeapType stringToHeapType(const char* str, bool prefix = false);
+  Type elementToType(Element& s);
   Type stringToLaneType(const char* str);
   bool isType(cashew::IString str) {
-    return stringToType(str, true) != none;
+    return stringToType(str, true) != Type::none;
   }
 
 public:
-  Expression* parseExpression(Element* s) {
-    return parseExpression(*s);
-  }
+  Expression* parseExpression(Element* s) { return parseExpression(*s); }
   Expression* parseExpression(Element& s);
 
-  MixedArena& getAllocator() {
-    return allocator;
-  }
+  Module& getModule() { return wasm; }
 
 private:
   Expression* makeExpression(Element& s);
@@ -175,37 +183,44 @@ private:
   Expression* makeUnary(Element& s, UnaryOp op);
   Expression* makeSelect(Element& s);
   Expression* makeDrop(Element& s);
-  Expression* makeHost(Element& s, HostOp op);
+  Expression* makeMemorySize(Element& s);
+  Expression* makeMemoryGrow(Element& s);
   Index getLocalIndex(Element& s);
-  Expression* makeGetLocal(Element& s);
-  Expression* makeTeeLocal(Element& s);
-  Expression* makeSetLocal(Element& s);
-  Expression* makeGetGlobal(Element& s);
-  Expression* makeSetGlobal(Element& s);
+  Expression* makeLocalGet(Element& s);
+  Expression* makeLocalTee(Element& s);
+  Expression* makeLocalSet(Element& s);
+  Expression* makeGlobalGet(Element& s);
+  Expression* makeGlobalSet(Element& s);
   Expression* makeBlock(Element& s);
   Expression* makeThenOrElse(Element& s);
   Expression* makeConst(Element& s, Type type);
   Expression* makeLoad(Element& s, Type type, bool isAtomic);
   Expression* makeStore(Element& s, Type type, bool isAtomic);
   Expression* makeAtomicRMWOrCmpxchg(Element& s, Type type);
-  Expression* makeAtomicRMW(Element& s, Type type, uint8_t bytes, const char* extra);
-  Expression* makeAtomicCmpxchg(Element& s, Type type, uint8_t bytes, const char* extra);
+  Expression*
+  makeAtomicRMW(Element& s, Type type, uint8_t bytes, const char* extra);
+  Expression*
+  makeAtomicCmpxchg(Element& s, Type type, uint8_t bytes, const char* extra);
   Expression* makeAtomicWait(Element& s, Type type);
   Expression* makeAtomicNotify(Element& s);
+  Expression* makeAtomicFence(Element& s);
   Expression* makeSIMDExtract(Element& s, SIMDExtractOp op, size_t lanes);
   Expression* makeSIMDReplace(Element& s, SIMDReplaceOp op, size_t lanes);
   Expression* makeSIMDShuffle(Element& s);
-  Expression* makeSIMDBitselect(Element& s);
+  Expression* makeSIMDTernary(Element& s, SIMDTernaryOp op);
   Expression* makeSIMDShift(Element& s, SIMDShiftOp op);
+  Expression* makeSIMDLoad(Element& s, SIMDLoadOp op);
   Expression* makeMemoryInit(Element& s);
   Expression* makeDataDrop(Element& s);
   Expression* makeMemoryCopy(Element& s);
   Expression* makeMemoryFill(Element& s);
+  Expression* makePush(Element& s);
+  Expression* makePop(Element& s);
   Expression* makeIf(Element& s);
   Expression* makeMaybeBlock(Element& s, size_t i, Type type);
   Expression* makeLoop(Element& s);
-  Expression* makeCall(Element& s);
-  Expression* makeCallIndirect(Element& s);
+  Expression* makeCall(Element& s, bool isReturn);
+  Expression* makeCallIndirect(Element& s, bool isReturn);
   template<class T>
   void parseCallOperands(Element& s, Index i, Index j, T* call) {
     while (i < j) {
@@ -217,9 +232,48 @@ private:
   Expression* makeBreak(Element& s);
   Expression* makeBreakTable(Element& s);
   Expression* makeReturn(Element& s);
+  Expression* makeRefNull(Element& s);
+  Expression* makeRefIsNull(Element& s);
+  Expression* makeRefFunc(Element& s);
+  Expression* makeRefEq(Element& s);
+  Expression* makeTry(Element& s);
+  Expression* makeTryOrCatchBody(Element& s, Type type, bool isTry);
+  Expression* makeThrow(Element& s);
+  Expression* makeRethrow(Element& s);
+  Expression* makeBrOnExn(Element& s);
+  Expression* makeTupleMake(Element& s);
+  Expression* makeTupleExtract(Element& s);
+  Expression* makeI31New(Element& s);
+  Expression* makeI31Get(Element& s, bool signed_);
+  Expression* makeRefTest(Element& s);
+  Expression* makeRefCast(Element& s);
+  Expression* makeBrOnCast(Element& s);
+  Expression* makeRttCanon(Element& s);
+  Expression* makeRttSub(Element& s);
+  Expression* makeStructNew(Element& s, bool default_);
+  Expression* makeStructGet(Element& s);
+  Expression* makeStructGet(Element& s, bool signed_);
+  Expression* makeStructSet(Element& s);
+  Expression* makeArrayNew(Element& s, bool default_);
+  Expression* makeArrayGet(Element& s);
+  Expression* makeArrayGet(Element& s, bool signed_);
+  Expression* makeArraySet(Element& s);
+  Expression* makeArrayLen(Element& s);
 
+  // Helper functions
   Type parseOptionalResultType(Element& s, Index& i);
   Index parseMemoryLimits(Element& s, Index i);
+  Index parseMemoryIndex(Element& s, Index i);
+  std::vector<Type> parseParamOrLocal(Element& s);
+  std::vector<NameType> parseParamOrLocal(Element& s, size_t& localIndex);
+  std::vector<Type> parseResults(Element& s);
+  Signature parseTypeRef(Element& s);
+  size_t parseTypeUse(Element& s,
+                      size_t startPos,
+                      Signature& functionSignature,
+                      std::vector<NameType>& namedParams);
+  size_t
+  parseTypeUse(Element& s, size_t startPos, Signature& functionSignature);
 
   void stringToBinary(const char* input, size_t size, std::vector<char>& data);
   void parseMemory(Element& s, bool preParseImport = false);
@@ -232,6 +286,7 @@ private:
   void parseElem(Element& s);
   void parseInnerElem(Element& s, Index i = 1, Expression* offset = nullptr);
   void parseType(Element& s);
+  void parseEvent(Element& s, bool preParseImport = false);
 
   Function::DebugLocation getDebugLocation(const SourceLocation& loc);
 };
