@@ -71,18 +71,35 @@ template<typename SubType, typename ReturnType = void> struct Visitor {
   ReturnType visitSelect(Select* curr) { return ReturnType(); }
   ReturnType visitDrop(Drop* curr) { return ReturnType(); }
   ReturnType visitReturn(Return* curr) { return ReturnType(); }
-  ReturnType visitHost(Host* curr) { return ReturnType(); }
+  ReturnType visitMemorySize(MemorySize* curr) { return ReturnType(); }
+  ReturnType visitMemoryGrow(MemoryGrow* curr) { return ReturnType(); }
   ReturnType visitRefNull(RefNull* curr) { return ReturnType(); }
   ReturnType visitRefIsNull(RefIsNull* curr) { return ReturnType(); }
   ReturnType visitRefFunc(RefFunc* curr) { return ReturnType(); }
+  ReturnType visitRefEq(RefEq* curr) { return ReturnType(); }
   ReturnType visitTry(Try* curr) { return ReturnType(); }
   ReturnType visitThrow(Throw* curr) { return ReturnType(); }
   ReturnType visitRethrow(Rethrow* curr) { return ReturnType(); }
   ReturnType visitBrOnExn(BrOnExn* curr) { return ReturnType(); }
   ReturnType visitNop(Nop* curr) { return ReturnType(); }
   ReturnType visitUnreachable(Unreachable* curr) { return ReturnType(); }
-  ReturnType visitPush(Push* curr) { return ReturnType(); }
   ReturnType visitPop(Pop* curr) { return ReturnType(); }
+  ReturnType visitTupleMake(TupleMake* curr) { return ReturnType(); }
+  ReturnType visitTupleExtract(TupleExtract* curr) { return ReturnType(); }
+  ReturnType visitI31New(I31New* curr) { return ReturnType(); }
+  ReturnType visitI31Get(I31Get* curr) { return ReturnType(); }
+  ReturnType visitRefTest(RefTest* curr) { return ReturnType(); }
+  ReturnType visitRefCast(RefCast* curr) { return ReturnType(); }
+  ReturnType visitBrOnCast(BrOnCast* curr) { return ReturnType(); }
+  ReturnType visitRttCanon(RttCanon* curr) { return ReturnType(); }
+  ReturnType visitRttSub(RttSub* curr) { return ReturnType(); }
+  ReturnType visitStructNew(StructNew* curr) { return ReturnType(); }
+  ReturnType visitStructGet(StructGet* curr) { return ReturnType(); }
+  ReturnType visitStructSet(StructSet* curr) { return ReturnType(); }
+  ReturnType visitArrayNew(ArrayNew* curr) { return ReturnType(); }
+  ReturnType visitArrayGet(ArrayGet* curr) { return ReturnType(); }
+  ReturnType visitArraySet(ArraySet* curr) { return ReturnType(); }
+  ReturnType visitArrayLen(ArrayLen* curr) { return ReturnType(); }
   // Module-level visitors
   ReturnType visitExport(Export* curr) { return ReturnType(); }
   ReturnType visitGlobal(Global* curr) { return ReturnType(); }
@@ -168,14 +185,18 @@ template<typename SubType, typename ReturnType = void> struct Visitor {
         DELEGATE(Drop);
       case Expression::Id::ReturnId:
         DELEGATE(Return);
-      case Expression::Id::HostId:
-        DELEGATE(Host);
+      case Expression::Id::MemorySizeId:
+        DELEGATE(MemorySize);
+      case Expression::Id::MemoryGrowId:
+        DELEGATE(MemoryGrow);
       case Expression::Id::RefNullId:
         DELEGATE(RefNull);
       case Expression::Id::RefIsNullId:
         DELEGATE(RefIsNull);
       case Expression::Id::RefFuncId:
         DELEGATE(RefFunc);
+      case Expression::Id::RefEqId:
+        DELEGATE(RefEq);
       case Expression::Id::TryId:
         DELEGATE(Try);
       case Expression::Id::ThrowId:
@@ -188,10 +209,40 @@ template<typename SubType, typename ReturnType = void> struct Visitor {
         DELEGATE(Nop);
       case Expression::Id::UnreachableId:
         DELEGATE(Unreachable);
-      case Expression::Id::PushId:
-        DELEGATE(Push);
       case Expression::Id::PopId:
         DELEGATE(Pop);
+      case Expression::Id::TupleMakeId:
+        DELEGATE(TupleMake);
+      case Expression::Id::TupleExtractId:
+        DELEGATE(TupleExtract);
+      case Expression::Id::I31NewId:
+        DELEGATE(I31New);
+      case Expression::Id::I31GetId:
+        DELEGATE(I31Get);
+      case Expression::Id::RefTestId:
+        DELEGATE(RefTest);
+      case Expression::Id::RefCastId:
+        DELEGATE(RefCast);
+      case Expression::Id::BrOnCastId:
+        DELEGATE(BrOnCast);
+      case Expression::Id::RttCanonId:
+        DELEGATE(RttCanon);
+      case Expression::Id::RttSubId:
+        DELEGATE(RttSub);
+      case Expression::Id::StructNewId:
+        DELEGATE(StructNew);
+      case Expression::Id::StructGetId:
+        DELEGATE(StructGet);
+      case Expression::Id::StructSetId:
+        DELEGATE(StructSet);
+      case Expression::Id::ArrayNewId:
+        DELEGATE(ArrayNew);
+      case Expression::Id::ArrayGetId:
+        DELEGATE(ArrayGet);
+      case Expression::Id::ArraySetId:
+        DELEGATE(ArraySet);
+      case Expression::Id::ArrayLenId:
+        DELEGATE(ArrayLen);
       case Expression::Id::InvalidId:
       default:
         WASM_UNREACHABLE("unexpected expression type");
@@ -249,18 +300,35 @@ struct OverriddenVisitor {
   UNIMPLEMENTED(Select);
   UNIMPLEMENTED(Drop);
   UNIMPLEMENTED(Return);
-  UNIMPLEMENTED(Host);
+  UNIMPLEMENTED(MemorySize);
+  UNIMPLEMENTED(MemoryGrow);
   UNIMPLEMENTED(RefNull);
   UNIMPLEMENTED(RefIsNull);
   UNIMPLEMENTED(RefFunc);
+  UNIMPLEMENTED(RefEq);
   UNIMPLEMENTED(Try);
   UNIMPLEMENTED(Throw);
   UNIMPLEMENTED(Rethrow);
   UNIMPLEMENTED(BrOnExn);
   UNIMPLEMENTED(Nop);
   UNIMPLEMENTED(Unreachable);
-  UNIMPLEMENTED(Push);
   UNIMPLEMENTED(Pop);
+  UNIMPLEMENTED(TupleMake);
+  UNIMPLEMENTED(TupleExtract);
+  UNIMPLEMENTED(I31New);
+  UNIMPLEMENTED(I31Get);
+  UNIMPLEMENTED(RefTest);
+  UNIMPLEMENTED(RefCast);
+  UNIMPLEMENTED(BrOnCast);
+  UNIMPLEMENTED(RttCanon);
+  UNIMPLEMENTED(RttSub);
+  UNIMPLEMENTED(StructNew);
+  UNIMPLEMENTED(StructGet);
+  UNIMPLEMENTED(StructSet);
+  UNIMPLEMENTED(ArrayNew);
+  UNIMPLEMENTED(ArrayGet);
+  UNIMPLEMENTED(ArraySet);
+  UNIMPLEMENTED(ArrayLen);
   UNIMPLEMENTED(Export);
   UNIMPLEMENTED(Global);
   UNIMPLEMENTED(Function);
@@ -347,14 +415,18 @@ struct OverriddenVisitor {
         DELEGATE(Drop);
       case Expression::Id::ReturnId:
         DELEGATE(Return);
-      case Expression::Id::HostId:
-        DELEGATE(Host);
+      case Expression::Id::MemorySizeId:
+        DELEGATE(MemorySize);
+      case Expression::Id::MemoryGrowId:
+        DELEGATE(MemoryGrow);
       case Expression::Id::RefNullId:
         DELEGATE(RefNull);
       case Expression::Id::RefIsNullId:
         DELEGATE(RefIsNull);
       case Expression::Id::RefFuncId:
         DELEGATE(RefFunc);
+      case Expression::Id::RefEqId:
+        DELEGATE(RefEq);
       case Expression::Id::TryId:
         DELEGATE(Try);
       case Expression::Id::ThrowId:
@@ -367,10 +439,40 @@ struct OverriddenVisitor {
         DELEGATE(Nop);
       case Expression::Id::UnreachableId:
         DELEGATE(Unreachable);
-      case Expression::Id::PushId:
-        DELEGATE(Push);
       case Expression::Id::PopId:
         DELEGATE(Pop);
+      case Expression::Id::TupleMakeId:
+        DELEGATE(TupleMake);
+      case Expression::Id::TupleExtractId:
+        DELEGATE(TupleExtract);
+      case Expression::Id::I31NewId:
+        DELEGATE(I31New);
+      case Expression::Id::I31GetId:
+        DELEGATE(I31Get);
+      case Expression::Id::RefTestId:
+        DELEGATE(RefTest);
+      case Expression::Id::RefCastId:
+        DELEGATE(RefCast);
+      case Expression::Id::BrOnCastId:
+        DELEGATE(BrOnCast);
+      case Expression::Id::RttCanonId:
+        DELEGATE(RttCanon);
+      case Expression::Id::RttSubId:
+        DELEGATE(RttSub);
+      case Expression::Id::StructNewId:
+        DELEGATE(StructNew);
+      case Expression::Id::StructGetId:
+        DELEGATE(StructGet);
+      case Expression::Id::StructSetId:
+        DELEGATE(StructSet);
+      case Expression::Id::ArrayNewId:
+        DELEGATE(ArrayNew);
+      case Expression::Id::ArrayGetId:
+        DELEGATE(ArrayGet);
+      case Expression::Id::ArraySetId:
+        DELEGATE(ArraySet);
+      case Expression::Id::ArrayLenId:
+        DELEGATE(ArrayLen);
       case Expression::Id::InvalidId:
       default:
         WASM_UNREACHABLE("unexpected expression type");
@@ -491,7 +593,10 @@ struct UnifiedExpressionVisitor : public Visitor<SubType, ReturnType> {
   ReturnType visitReturn(Return* curr) {
     return static_cast<SubType*>(this)->visitExpression(curr);
   }
-  ReturnType visitHost(Host* curr) {
+  ReturnType visitMemorySize(MemorySize* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitMemoryGrow(MemoryGrow* curr) {
     return static_cast<SubType*>(this)->visitExpression(curr);
   }
   ReturnType visitRefNull(RefNull* curr) {
@@ -503,6 +608,9 @@ struct UnifiedExpressionVisitor : public Visitor<SubType, ReturnType> {
   ReturnType visitRefFunc(RefFunc* curr) {
     return static_cast<SubType*>(this)->visitExpression(curr);
   }
+  ReturnType visitRefEq(RefEq* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
   ReturnType visitTry(Try* curr) {
     return static_cast<SubType*>(this)->visitExpression(curr);
   }
@@ -521,10 +629,55 @@ struct UnifiedExpressionVisitor : public Visitor<SubType, ReturnType> {
   ReturnType visitUnreachable(Unreachable* curr) {
     return static_cast<SubType*>(this)->visitExpression(curr);
   }
-  ReturnType visitPush(Push* curr) {
+  ReturnType visitPop(Pop* curr) {
     return static_cast<SubType*>(this)->visitExpression(curr);
   }
-  ReturnType visitPop(Pop* curr) {
+  ReturnType visitTupleMake(TupleMake* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitTupleExtract(TupleExtract* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitI31New(I31New* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitI31Get(I31Get* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitRefTest(RefTest* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitRefCast(RefCast* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitBrOnCast(BrOnCast* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitRttCanon(RttCanon* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitRttSub(RttSub* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitStructNew(StructNew* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitStructGet(StructGet* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitStructSet(StructSet* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitArrayNew(ArrayNew* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitArrayGet(ArrayGet* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitArraySet(ArraySet* curr) {
+    return static_cast<SubType*>(this)->visitExpression(curr);
+  }
+  ReturnType visitArrayLen(ArrayLen* curr) {
     return static_cast<SubType*>(this)->visitExpression(curr);
   }
 };
@@ -802,8 +955,11 @@ struct Walker : public VisitorType {
   static void doVisitReturn(SubType* self, Expression** currp) {
     self->visitReturn((*currp)->cast<Return>());
   }
-  static void doVisitHost(SubType* self, Expression** currp) {
-    self->visitHost((*currp)->cast<Host>());
+  static void doVisitMemorySize(SubType* self, Expression** currp) {
+    self->visitMemorySize((*currp)->cast<MemorySize>());
+  }
+  static void doVisitMemoryGrow(SubType* self, Expression** currp) {
+    self->visitMemoryGrow((*currp)->cast<MemoryGrow>());
   }
   static void doVisitRefNull(SubType* self, Expression** currp) {
     self->visitRefNull((*currp)->cast<RefNull>());
@@ -814,6 +970,9 @@ struct Walker : public VisitorType {
   static void doVisitRefFunc(SubType* self, Expression** currp) {
     self->visitRefFunc((*currp)->cast<RefFunc>());
   }
+  static void doVisitRefEq(SubType* self, Expression** currp) {
+    self->visitRefEq((*currp)->cast<RefEq>());
+  }
   static void doVisitTry(SubType* self, Expression** currp) {
     self->visitTry((*currp)->cast<Try>());
   }
@@ -832,12 +991,57 @@ struct Walker : public VisitorType {
   static void doVisitUnreachable(SubType* self, Expression** currp) {
     self->visitUnreachable((*currp)->cast<Unreachable>());
   }
-  static void doVisitPush(SubType* self, Expression** currp) {
-    self->visitPush((*currp)->cast<Push>());
-  }
   static void doVisitPop(SubType* self, Expression** currp) {
     self->visitPop((*currp)->cast<Pop>());
   }
+  static void doVisitTupleMake(SubType* self, Expression** currp) {
+    self->visitTupleMake((*currp)->cast<TupleMake>());
+  }
+  static void doVisitTupleExtract(SubType* self, Expression** currp) {
+    self->visitTupleExtract((*currp)->cast<TupleExtract>());
+  }
+  static void doVisitI31New(SubType* self, Expression** currp) {
+    self->visitI31New((*currp)->cast<I31New>());
+  }
+  static void doVisitI31Get(SubType* self, Expression** currp) {
+    self->visitI31Get((*currp)->cast<I31Get>());
+  }
+  static void doVisitRefTest(SubType* self, Expression** currp) {
+    self->visitRefTest((*currp)->cast<RefTest>());
+  }
+  static void doVisitRefCast(SubType* self, Expression** currp) {
+    self->visitRefCast((*currp)->cast<RefCast>());
+  }
+  static void doVisitBrOnCast(SubType* self, Expression** currp) {
+    self->visitBrOnCast((*currp)->cast<BrOnCast>());
+  }
+  static void doVisitRttCanon(SubType* self, Expression** currp) {
+    self->visitRttCanon((*currp)->cast<RttCanon>());
+  }
+  static void doVisitRttSub(SubType* self, Expression** currp) {
+    self->visitRttSub((*currp)->cast<RttSub>());
+  }
+  static void doVisitStructNew(SubType* self, Expression** currp) {
+    self->visitStructNew((*currp)->cast<StructNew>());
+  }
+  static void doVisitStructGet(SubType* self, Expression** currp) {
+    self->visitStructGet((*currp)->cast<StructGet>());
+  }
+  static void doVisitStructSet(SubType* self, Expression** currp) {
+    self->visitStructSet((*currp)->cast<StructSet>());
+  }
+  static void doVisitArrayNew(SubType* self, Expression** currp) {
+    self->visitArrayNew((*currp)->cast<ArrayNew>());
+  }
+  static void doVisitArrayGet(SubType* self, Expression** currp) {
+    self->visitArrayGet((*currp)->cast<ArrayGet>());
+  }
+  static void doVisitArraySet(SubType* self, Expression** currp) {
+    self->visitArraySet((*currp)->cast<ArraySet>());
+  }
+  static void doVisitArrayLen(SubType* self, Expression** currp) {
+    self->visitArrayLen((*currp)->cast<ArrayLen>());
+  }
 
   void setModule(Module* module) { currModule = module; }
 
@@ -1064,14 +1268,13 @@ struct PostWalker : public Walker<SubType, VisitorType> {
         self->maybePushTask(SubType::scan, &curr->cast<Return>()->value);
         break;
       }
-      case Expression::Id::HostId: {
-        self->pushTask(SubType::doVisitHost, currp);
-        auto& list = curr->cast<Host>()->operands;
-        for (int i = int(list.size()) - 1; i >= 0; i--) {
-          self->pushTask(SubType::scan, &list[i]);
-        }
+      case Expression::Id::MemorySizeId:
+        self->pushTask(SubType::doVisitMemorySize, currp);
+        break;
+      case Expression::Id::MemoryGrowId:
+        self->pushTask(SubType::doVisitMemoryGrow, currp);
+        self->pushTask(SubType::scan, &curr->cast<MemoryGrow>()->delta);
         break;
-      }
       case Expression::Id::RefNullId: {
         self->pushTask(SubType::doVisitRefNull, currp);
         break;
@@ -1085,6 +1288,12 @@ struct PostWalker : public Walker<SubType, VisitorType> {
         self->pushTask(SubType::doVisitRefFunc, currp);
         break;
       }
+      case Expression::Id::RefEqId: {
+        self->pushTask(SubType::doVisitRefEq, currp);
+        self->pushTask(SubType::scan, &curr->cast<RefEq>()->right);
+        self->pushTask(SubType::scan, &curr->cast<RefEq>()->left);
+        break;
+      }
       case Expression::Id::TryId: {
         self->pushTask(SubType::doVisitTry, currp);
         self->pushTask(SubType::scan, &curr->cast<Try>()->catchBody);
@@ -1117,15 +1326,81 @@ struct PostWalker : public Walker<SubType, VisitorType> {
         self->pushTask(SubType::doVisitUnreachable, currp);
         break;
       }
-      case Expression::Id::PushId: {
-        self->pushTask(SubType::doVisitPush, currp);
-        self->pushTask(SubType::scan, &curr->cast<Push>()->value);
-        break;
-      }
       case Expression::Id::PopId: {
         self->pushTask(SubType::doVisitPop, currp);
         break;
       }
+      case Expression::Id::TupleMakeId: {
+        self->pushTask(SubType::doVisitTupleMake, currp);
+        auto& operands = curr->cast<TupleMake>()->operands;
+        for (int i = int(operands.size()) - 1; i >= 0; --i) {
+          self->pushTask(SubType::scan, &operands[i]);
+        }
+        break;
+      }
+      case Expression::Id::TupleExtractId: {
+        self->pushTask(SubType::doVisitTupleExtract, currp);
+        self->pushTask(SubType::scan, &curr->cast<TupleExtract>()->tuple);
+        break;
+      }
+      case Expression::Id::I31NewId: {
+        self->pushTask(SubType::doVisitI31New, currp);
+        self->pushTask(SubType::scan, &curr->cast<I31New>()->value);
+        break;
+      }
+      case Expression::Id::I31GetId: {
+        self->pushTask(SubType::doVisitI31Get, currp);
+        self->pushTask(SubType::scan, &curr->cast<I31Get>()->i31);
+        break;
+      }
+      case Expression::Id::RefTestId:
+        self->pushTask(SubType::doVisitRefTest, currp);
+        WASM_UNREACHABLE("TODO (gc): ref.test");
+        break;
+      case Expression::Id::RefCastId:
+        self->pushTask(SubType::doVisitRefCast, currp);
+        WASM_UNREACHABLE("TODO (gc): ref.cast");
+        break;
+      case Expression::Id::BrOnCastId:
+        self->pushTask(SubType::doVisitBrOnCast, currp);
+        WASM_UNREACHABLE("TODO (gc): br_on_cast");
+        break;
+      case Expression::Id::RttCanonId:
+        self->pushTask(SubType::doVisitRttCanon, currp);
+        WASM_UNREACHABLE("TODO (gc): rtt.canon");
+        break;
+      case Expression::Id::RttSubId:
+        self->pushTask(SubType::doVisitRttSub, currp);
+        WASM_UNREACHABLE("TODO (gc): rtt.sub");
+        break;
+      case Expression::Id::StructNewId:
+        self->pushTask(SubType::doVisitStructNew, currp);
+        WASM_UNREACHABLE("TODO (gc): struct.new");
+        break;
+      case Expression::Id::StructGetId:
+        self->pushTask(SubType::doVisitStructGet, currp);
+        WASM_UNREACHABLE("TODO (gc): struct.get");
+        break;
+      case Expression::Id::StructSetId:
+        self->pushTask(SubType::doVisitStructSet, currp);
+        WASM_UNREACHABLE("TODO (gc): struct.set");
+        break;
+      case Expression::Id::ArrayNewId:
+        self->pushTask(SubType::doVisitArrayNew, currp);
+        WASM_UNREACHABLE("TODO (gc): array.new");
+        break;
+      case Expression::Id::ArrayGetId:
+        self->pushTask(SubType::doVisitArrayGet, currp);
+        WASM_UNREACHABLE("TODO (gc): array.get");
+        break;
+      case Expression::Id::ArraySetId:
+        self->pushTask(SubType::doVisitArraySet, currp);
+        WASM_UNREACHABLE("TODO (gc): array.set");
+        break;
+      case Expression::Id::ArrayLenId:
+        self->pushTask(SubType::doVisitArrayLen, currp);
+        WASM_UNREACHABLE("TODO (gc): array.len");
+        break;
       case Expression::Id::NumExpressionIds:
         WASM_UNREACHABLE("unexpected expression type");
     }
@@ -1159,7 +1434,7 @@ struct ControlFlowWalker : public PostWalker<SubType, VisitorType> {
           return curr;
         }
       } else {
-        // an if, ignorable
+        // an if or try, ignorable
         assert(curr->template is<If>() || curr->template is<Try>());
       }
       if (i == 0) {
@@ -1190,7 +1465,8 @@ struct ControlFlowWalker : public PostWalker<SubType, VisitorType> {
         self->pushTask(SubType::doPostVisitControlFlow, currp);
         break;
       }
-      default: {}
+      default: {
+      }
     }
 
     PostWalker<SubType, VisitorType>::scan(self, currp);
@@ -1203,7 +1479,8 @@ struct ControlFlowWalker : public PostWalker<SubType, VisitorType> {
         self->pushTask(SubType::doPreVisitControlFlow, currp);
         break;
       }
-      default: {}
+      default: {
+      }
     }
   }
 };
