@@ -14,10 +14,6 @@ void BinaryenShimDisposeBinaryenModuleAllocateAndWriteResult(
     BinaryenModuleAllocateAndWriteResult result
 );
 
-void BinaryenModuleOptimizeWithSettings(
-    BinaryenModuleRef module, int shrinkLevel, int optimizeLevel, int debugInfo
-);
-
 void BinaryenModuleRunPassesWithSettings(
     BinaryenModuleRef module, const char** passes, BinaryenIndex numPasses,
     int shrinkLevel, int optimizeLevel, int debugInfo
